/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenixpro/configs/Configs.hpp"
#include "ctre/phoenixpro/controls/ControlRequest.hpp"
#include "ctre/phoenixpro/networking/interfaces/Control_Interface.h"
#include <sstream>
#include <units/angle.h>
#include <units/dimensionless.h>
#include <units/frequency.h>
#include <units/time.h>


namespace ctre {
namespace phoenixpro {
namespace controls {

/**
 * Request PID to target position with duty-cycle feed forward.
 *
 * This control mode will set the motor's position setpoint to the position specified by the user. In
 * addition, it will apply an additional voltage as an arbitrary feed-forward value.
 */
class MotionMagicDutyCycle : public ControlRequest
{
    bool ApplyConfigsOnRequest;

    ctre::phoenix::StatusCode SendRequest(const char *network, uint32_t deviceHash, bool cancelOtherRequests, std::shared_ptr<ControlRequest> &req)
    {
        std::stringstream ss;
        auto& ref = requestReference.GetNameValues();
        ss << Position.to<double>(); ref["Position"] = ss.str(); ss.str(std::string());
        ss << EnableFOC; ref["EnableFOC"] = ss.str(); ss.str(std::string());
        ss << FeedForward.to<double>(); ref["FeedForward"] = ss.str(); ss.str(std::string());
        ss << Slot; ref["Slot"] = ss.str(); ss.str(std::string());
        ss << OverrideBrakeDurNeutral; ref["OverrideBrakeDurNeutral"] = ss.str(); ss.str(std::string());

        if (req.get() != this)
        {
            auto const reqCast = dynamic_cast<MotionMagicDutyCycle *>(req.get());
            if (reqCast != nullptr)
            {
                *reqCast = *this;
            }
            else
            {
                req = std::make_shared<MotionMagicDutyCycle>(*this);
            }
        }

        
        std::string strs{ss.str()};
        c_ctre_phoenixpro_requestConfigApply(network, deviceHash, ConfigTimeout.to<double>(), strs.c_str(), strs.length(), ApplyConfigsOnRequest);
        ApplyConfigsOnRequest = false;
        return c_ctre_phoenixpro_RequestControlMotionMagicDutyCycle(network, deviceHash, UpdateFreqHz.to<double>(), cancelOtherRequests, Position.to<double>(), EnableFOC, FeedForward.to<double>(), Slot, OverrideBrakeDurNeutral);
    }

public:
    /**
     * Position to drive toward in rotations.
     */
    units::angle::turn_t Position;
    /**
     * Set to true to use FOC commutation, which increases peak power by ~15%. Set
     * to false to use trapezoidal commutation.  FOC improves motor performance by
     * leveraging torque (current) control.  However, this may be inconvenient for
     * applications that require specifying duty cycle or voltage.  CTR-Electronics
     * has developed a hybrid method that combines the performances gains of FOC
     * while still allowing applications to provide duty-cycle or voltage demand. 
     * This not to be confused with simple sinusoidal control or phase voltage
     * control which lacks the performance gains.
     */
    bool EnableFOC;
    /**
     * Feedforward to apply in fractional units between -1 and +1.
     */
    units::dimensionless::scalar_t FeedForward;
    /**
     * Select which gains are applied by selecting the slot.  Use the configuration
     * api to set the gain values for the selected slot before enabling this
     * feature. Slot must be within [0,2].
     */
    int Slot;
    /**
     * Set to true to static-brake the rotor when output is zero (or within
     * deadband).  Set to false to use the NeutralMode configuration setting
     * (default). This flag exists to provide the fundamental behavior of this
     * control when output is zero, which is to provide 0V to the motor.
     */
    bool OverrideBrakeDurNeutral;

    

    /**
     * \brief The timeout when sending configs associated with this control
     */
    units::time::second_t ConfigTimeout{0.1_s};

    /**
     * \brief The period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     */
    units::frequency::hertz_t UpdateFreqHz{100_Hz}; // Default to 100_Hz

    /**
     *\brief  Request PID to target position with duty-cycle feed forward.
     *
     *\details  This control mode will set the motor's position setpoint to the
     *          position specified by the user. In addition, it will apply an
     *          additional voltage as an arbitrary feed-forward value.
     *
     * \param Position     Position to drive toward in rotations.
     * \param EnableFOC     Set to true to use FOC commutation, which increases
     *                      peak power by ~15%. Set to false to use trapezoidal
     *                      commutation.  FOC improves motor performance by
     *                      leveraging torque (current) control.  However, this
     *                      may be inconvenient for applications that require
     *                      specifying duty cycle or voltage.  CTR-Electronics
     *                      has developed a hybrid method that combines the
     *                      performances gains of FOC while still allowing
     *                      applications to provide duty-cycle or voltage
     *                      demand.  This not to be confused with simple
     *                      sinusoidal control or phase voltage control which
     *                      lacks the performance gains.
     * \param FeedForward     Feedforward to apply in fractional units between
     *                        -1 and +1.
     * \param Slot     Select which gains are applied by selecting the slot. 
     *                 Use the configuration api to set the gain values for the
     *                 selected slot before enabling this feature. Slot must be
     *                 within [0,2].
     * \param OverrideBrakeDurNeutral     Set to true to static-brake the rotor
     *                                    when output is zero (or within
     *                                    deadband).  Set to false to use the
     *                                    NeutralMode configuration setting
     *                                    (default). This flag exists to provide
     *                                    the fundamental behavior of this
     *                                    control when output is zero, which is
     *                                    to provide 0V to the motor.
     */
    MotionMagicDutyCycle(units::angle::turn_t Position, bool EnableFOC, units::dimensionless::scalar_t FeedForward, int Slot, bool OverrideBrakeDurNeutral) : ControlRequest{"MotionMagicDutyCycle"}, ApplyConfigsOnRequest{false}
    {
        this->Position = Position;
        this->EnableFOC = EnableFOC;
        this->FeedForward = FeedForward;
        this->Slot = Slot;
        this->OverrideBrakeDurNeutral = OverrideBrakeDurNeutral;
    }

    /**
     *\brief  Request PID to target position with duty-cycle feed forward.
     *
     *\details  This control mode will set the motor's position setpoint to the
     *          position specified by the user. In addition, it will apply an
     *          additional voltage as an arbitrary feed-forward value.
     *
     * \param Position     Position to drive toward in rotations.
     */
    MotionMagicDutyCycle(units::angle::turn_t Position) : MotionMagicDutyCycle{Position, true, 0.0, 0, false}
    {}
    
    /**
     * \brief Modifies this Control Request's Position parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newPosition Parameter to modify
     * \returns Itself
     */
    MotionMagicDutyCycle& WithPosition(units::angle::turn_t newPosition)
    {
        Position = newPosition;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's EnableFOC parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newEnableFOC Parameter to modify
     * \returns Itself
     */
    MotionMagicDutyCycle& WithEnableFOC(bool newEnableFOC)
    {
        EnableFOC = newEnableFOC;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's FeedForward parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newFeedForward Parameter to modify
     * \returns Itself
     */
    MotionMagicDutyCycle& WithFeedForward(units::dimensionless::scalar_t newFeedForward)
    {
        FeedForward = newFeedForward;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's Slot parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newSlot Parameter to modify
     * \returns Itself
     */
    MotionMagicDutyCycle& WithSlot(int newSlot)
    {
        Slot = newSlot;
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's OverrideBrakeDurNeutral parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newOverrideBrakeDurNeutral Parameter to modify
     * \returns Itself
     */
    MotionMagicDutyCycle& WithOverrideBrakeDurNeutral(bool newOverrideBrakeDurNeutral)
    {
        OverrideBrakeDurNeutral = newOverrideBrakeDurNeutral;
        return *this;
    }
    /**
     * \brief Sets the period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     *
     * \param newUpdateFreqHz Parameter to modify
     * \returns Itself
     */
    MotionMagicDutyCycle &WithUpdateFreqHz(units::frequency::hertz_t newUpdateFreqHz)
    {
        UpdateFreqHz = newUpdateFreqHz;
        return *this;
    }
    /**
     * Returns a string representation of the object.
     *
     * \returns a string representation of the object.
     */
    std::string ToString() const
    {
        std::stringstream ss;
        ss << "class: MotionMagicDutyCycle" << std::endl;
        ss << "Position: " << Position.to<double>() << std::endl;
        ss << "EnableFOC: " << EnableFOC << std::endl;
        ss << "FeedForward: " << FeedForward.to<double>() << std::endl;
        ss << "Slot: " << Slot << std::endl;
        ss << "OverrideBrakeDurNeutral: " << OverrideBrakeDurNeutral << std::endl;
        return ss.str();
    }

    /**
     * \brief Forces configs to be applied the next time this is used in a setControl.
     *        This is not necessary in the majority of cases, because Phoenix will make sure configs are
     *        properly set when they are not already set
     */
    void ForceApplyConfigs() { ApplyConfigsOnRequest = true; }
};

}
}
}

